/*****************************************************************************
   KXTJ3.cpp

   Copyright (c) 2021 ROHM Co.,Ltd.

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
 ******************************************************************************/
#include <Arduino.h>
#include <Wire.h>
#include <KXTJ3.h>

KXTJ3::KXTJ3(bool addr_pin)
{
    if (addr_pin) {
        Address = KXTJ3_DEVICE_ADDRESS_0F;
    } else {
        Address = KXTJ3_DEVICE_ADDRESS_0E;
    }

    return;
}

KXTJ3::~KXTJ3()
{
    Address = 0;

    return;
}

float32 KXTJ3::get_version(void)
{
    return KXTJ3_DRIVER_VERSION;
}

int32_t KXTJ3::init(void)
{
    int32_t result;
    uint8_t control1;

    result = init_setting_ic(&control1);
    if (result == KXTJ3_COMM_OK) {
        result = init_setting_gval(control1);
    }

    return result;
}

int32_t KXTJ3::start(void)
{
    int32_t result;
    uint8_t val;

    val    = RegCtlVal | KXTJ3_CNTL1_PC1;
    result = write(KXTJ3_CNTL1, &val, sizeof(val));

    return result;
}

int32_t KXTJ3::stop(void)
{
    int32_t result;
    uint8_t val;

    val    = RegCtlVal & ~KXTJ3_CNTL1_PC1;
    result = write(KXTJ3_CNTL1, &val, sizeof(val));

    return result;
}

int32_t KXTJ3::get_val(uint8_t *raw)
{
    int32_t result;

    result = read(KXTJ3_XOUT_L, raw, KXTJ3_READ_DATA_SIZE);

    return result;
}

int32_t KXTJ3::get_val(float32 *acc)
{
    int32_t result;
    uint8_t raw[KXTJ3_READ_DATA_SIZE];
    int16_t acc_temp;
    uint8_t cnt, cnt_temp;

    result = get_val(raw);
    if (result == KXTJ3_COMM_OK) {
        for (cnt = 0; cnt < KXTJ3_AXIS_SIZE; cnt++) {
            cnt_temp = 2 * cnt;
            acc_temp  = (int16_t)raw[cnt_temp];
            acc_temp |= (int16_t)raw[cnt_temp + 1] << 8;
            acc[cnt] = (float32)acc_temp / GDivide;
        }
    }

    return result;
}

int32_t KXTJ3::init_setting_ic(uint8_t *checkval)
{
    int32_t result;
    uint8_t val;

    result = read(KXTJ3_WHO_AM_I, &val, sizeof(val));
    if (result == KXTJ3_COMM_OK) {
        if (val == KXTJ3_WAI_VAL) {
            val = KXTJ3_CNTL1_VAL;
            (void)write(KXTJ3_CNTL1, &val, sizeof(val));

            val = KXTJ3_DATA_CNTL_VAL;
            (void)write(KXTJ3_DATA_CNTL, &val, sizeof(val));

            (void)read(KXTJ3_CNTL1, &val, sizeof(val));
            *checkval = val;
        } else {
            result = KXTJ3_WAI_ERROR;
        }
    }

    return result;
}

int32_t KXTJ3::init_setting_gval(uint8_t checkval)
{
    int32_t result = KXTJ3_COMM_OK;
    int32_t gsense;
    uint8_t res, gsel, en16g, odr;

    RegCtlVal = checkval;
    res       = checkval & KXTJ3_CNTL1_RES_MASK;
    gsel      = checkval & KXTJ3_CNTL1_GSELMASK;
    en16g     = checkval & KXTJ3_CNTL1_EN16GMASK;
    odr       = KXTJ3_DATA_CNTL_OSAA_CLEAR & KXTJ3_DATA_CNTL_VAL;

    if (res == KXTJ3_CNTL1_RES_LOWPOWER) {
        if ((gsel == KXTJ3_CNTL1_GSEL_14BIT) || (odr > KXTJ3_DATA_CNTL_ODRCHECK)) {
            result = KXTJ3_PARAM_ERROR;
        }
    }

    if (result == KXTJ3_COMM_OK) {
        if (en16g == KXTJ3_CNTL1_GSEL_16G) {
            gsense = KXTJ3_GSENS_16G;
        } else {
            switch (gsel) {
                case KXTJ3_CNTL1_GSEL_2G:
                    gsense = KXTJ3_GSENS_2G;
                    break;
                case KXTJ3_CNTL1_GSEL_4G:
                    gsense = KXTJ3_GSENS_4G;
                    break;
                default:
                    gsense = KXTJ3_GSENS_8G;
                    break;
                }
        }
        GDivide = ((int32_t)1 << KXTJ3_DIVIDE_SHIFT) / gsense;
    }

    return result;
}

int32_t KXTJ3::write(uint8_t reg, uint8_t *data, int32_t size)
{
    int32_t result;

    Wire.beginTransmission(Address);
    (void)Wire.write(reg);
    (void)Wire.write(data, (uint8_t)size);
    result = Wire.endTransmission(true);
    if (result == 0) {
        result = KXTJ3_COMM_OK;
    } else {
        result = KXTJ3_COMM_ERROR;
    }

    return result;
}


int32_t KXTJ3::read(uint8_t reg, uint8_t *data, int32_t size)
{
    int32_t result;
    uint8_t cnt;
    uint8_t receive_size;

    Wire.beginTransmission(Address);
    (void)Wire.write(reg);
    result = Wire.endTransmission(false);
    if (result == 0) {
        receive_size = Wire.requestFrom(Address, (uint8_t)size);
        if (receive_size == size) {
            for (cnt = 0; cnt < size; cnt++) {
                data[cnt] = Wire.read();
            }
            result = KXTJ3_COMM_OK;
        } else {
            result = KXTJ3_COMM_ERROR;
        }
    } else {
        result = KXTJ3_COMM_ERROR;
    }

    return result;
}